#!/usr/bin/perl

#GLOBAL variables
use vars qw/$totalbytes $readbytes $bitlen $minbitlen @abuff $bufflen @record @casrecord $irgtimer $reclen $longzero $INFILE $CASFILE $LOGFILE/;

use Getopt::Long qw(:config pass_through permute);
use IO::Handle;
use Pod::Usage;

use constant {
    ZEROBIT => chr(0),
    ONEBIT => chr(255),
    };

use constant {
    START => ZEROBIT,	#start bit is 0
    STOP => ONEBIT,	#stop bit is 1
    MINBITLEN => 22,
    MARK => ONEBIT,
    };

use constant {
    ZERO_DETECTED => -5,
    END_OF_FILE => -1,
    NOT_GOOD => -4,
    IRG_DETECTED => -2,
    };

# Subroutines

sub fillbuffer {

# This subroutine, among other benefits, allows reading input data from STDIN

    my $inbuff;
    my $got;
    my $mult=1;		#used to read 2MB during firs call

    $mult=2 if $bufflen==0;
    $got=sysread($INFILE,$inbuff,1024*1024*$mult);
    return END_OF_FILE if (!$got);
    $totalbytes+=$got;
    if ($bufflen > 0) {
	if (@_[0]!=$readbytes) {
	    $readbytes-=(1024*1024);
	}
	@_[0]-=(1024*1024);
    }
    splice(@abuff,0,1024*1024);
    $bufflen=push(@abuff,split('',$inbuff));
# for debugging
#    print $LOGFILE "got: $got, bufflen: $bufflen, readbytes: $readbytes,  total: $totalbytes\n";
    return 1;
}

sub getbyte_probe {

# NOT USED
# possible alternative, simpler and faster way of detecting bits
# put in getbyte() between "Full scan" and "END Full scan" comments
# replacing current contents.
# Use on Your own risk ;)
# Not tested extensively

    my $byte;

    while($i++<8) {			#read 8 bits
	$readbytes+=int($bitlen/2);
	if ($readbytes >= $bufflen) {
	    return END_OF_FILE if fillbuffer($readbytes) < 0;
	}
	if ($abuff[$readbytes] eq ZEROBIT) {		# probe bit at half of bit width
	    $byte.='0';
	} else {
	    $byte.='1';
	}
	$readbytes+=$bitlen-(int($bitlen/2));
	if ($readbytes >= $bufflen) {
	    return END_OF_FILE if fillbuffer($readbytes) < 0;
	}
    }

}

sub getbyte {

# get one byte

    my $i=0;
    my $bit;
    my $savedbit;
    my $byte='';
    my $casbyte='';
    my $sum=0;
    my $result;
    my $bitcounter=0;
    my $temptimer=0;
    my $spikedetect;

    #skip until start bit
    while($abuff[$readbytes] ne START) {
	$readbytes++;
	if ($readbytes == $bufflen) {
	    return END_OF_FILE if fillbuffer($readbytes) < 0;
	}
    }
    #skip start bit
    #it should be $bitlen long
    $bitcounter=0;
    while ($bitcounter<$bitlen) {
	$bit=$abuff[$readbytes];	#it should be START, unless random spike
	while (($abuff[$readbytes] eq $bit) and ($bitcounter<$bitlen)) {
	    $readbytes++; $bitcounter++;
	    if ($readbytes == $bufflen) {
		return END_OF_FILE if fillbuffer($readbytes) < 0;
	    }
	}
    }
    if ($bitcounter==$bitlen) {
	#Look for "long zero" signal
	$temptimer=$readbytes;
	while ($abuff[$temptimer] eq ZEROBIT) {
	    $temptimer++;
	    if ($temptimer == $bufflen) {
		return END_OF_FILE if fillbuffer($temptimer) < 0;
	    }
	}
	if ($temptimer-$readbytes>=$bitlen*9) {		# found long zero
	    $longzero=$temptimer-$readbytes;
	    $readbytes=$temptimer;
	    return ZERO_DETECTED;
	}
    }

# FULL scan
    while($i++<8) {			#read 8 bits
	$bitcounter=0;
	while (true) {
	  while ($bitcounter<$minbitlen) {
	    $bit=$abuff[$readbytes];
	    while (($bitcounter<$bitlen) and ($abuff[$readbytes] eq $bit)) {
		$sum+=ord($abuff[$readbytes]);
		$readbytes++; $bitcounter++;
		if ($readbytes == $bufflen) {
		    last if fillbuffer($readbytes) < 0;
		}
	    }
	  }
	    last if $bitcounter>=$bitlen;
# Detection of random, narrow signal changes (spikes)
# This is a bit messy ;)
	    $temptimer=$readbytes;
	    $spikedetect=0;
	    $savedbit=$bit;
	    $bit=$abuff[$temptimer];
	    while (($abuff[$temptimer] eq $bit) and ($spikedetect<$minbitlen)) {
		$temptimer++; $spikedetect++;
		if ($temptimer == $bufflen) {
		    return END_OF_FILE if fillbuffer($temptimer) < 0;
		}
	    }
	    if ($spikedetect>=$minbitlen) {
		$sum+=($bitlen-$bitcounter)*ord($bit);	# there was "short read", fill up the bit.
		$readbytes+=($bitlen-$bitcounter);	# and correct the counter
		if ($readbytes == $bufflen) {
		    return END_OF_FILE if fillbuffer($readbytes) < 0;
		}
		last;
	    }  else {
		print $LOGFILE "It was a spike: $spikedetect in byte $reclen \n";
		$sum+=$spikedetect*ord($savedbit);	# add the one that shoud be here
		$readbytes=$temptimer;
	    }
	}
# Bit is detected
	if ($sum < $bitlen*255*0.3) { $byte.='0'; }		#less then 30% ones means zero
	elsif ($sum > $bitlen*255*0.7) { $byte.='1'; }		#more then 70% ones means one
	else {
#	    print "Doubtful byte. sum=$sum\n";
	    if ($sum < $bitlen*255/2) { $byte.='0'; } else { $byte.='1'; }	#50% chance of 0/1
	}
	$sum=0;
    }
# END Full scan
    $casbyte=pack("b8",$byte); #convert bin to char
    $byte=unpack("C", pack("b8",$byte)); #convert bin to text
    @_[0]=$byte;
    @_[1]=$casbyte;

    #skip until stop bit
    while($abuff[$readbytes] ne STOP) {
	$readbytes++;
	if ($readbytes == $bufflen) {
	    return END_OF_FILE if fillbuffer($readbytes) < 0;
	}
    }

    #skip stop bit
    #just look for start of start bit, which is directly after stop bit
    $bitcounter=0;
    $temptimer=$readbytes;
    while ($bitcounter<$minbitlen) {	# stop should not be shorter
	$bit=$abuff[$temptimer];	# this should be STOP
	while(($abuff[$temptimer] eq $bit) and ($bitcounter<=($bitlen*8))) {
	    $temptimer++; $bitcounter++;
	    if ($temptimer == $bufflen) {
		return END_OF_FILE if fillbuffer($temptimer) < 0;
	    }
	}

	return IRG_DETECTED if $bitcounter >= ($bitlen*8); # this is not stop bit, this is IRG

	#just in case, check for a spike in STOP/IRG
	$spikedetect=0;
	$bit=$abuff[$temptimer];
	while (($abuff[$temptimer] eq $bit) and ($spikedetect<$minbitlen)) {
	    $temptimer++; $spikedetect++; $bitcounter++;
	    if ($temptimer == $bufflen) {
		return END_OF_FILE if fillbuffer($temptimer) < 0;
	    }
	}
	if ($spikedetect>=$minbitlen) {$temptimer-=$spikedetect; last;
	 }  else { print $LOGFILE "spike in stop/start detected: $spikedetect\n"; $bitcounter=0; }
    }
    $readbytes=$temptimer;
    return 1;
}

sub readmarker {

# Detect marker at start of data block
# and compute bit lenght

    my $i=0;
    my $bit;
    my $count=1;	# sample counter
    my $temptimer;
    my $savedbitlen;
    my $start;

    print $LOGFILE "start readbytes: $readbytes\n";
    $start=$readbytes;
    return NOT_GOOD if ($abuff[$readbytes] ne START);	#it MUST start from 0 (start bit)
    while($i++<20) {				#20 marker bits
	$bit=$abuff[$readbytes];		#compare next samples to the first one
	while (($abuff[$readbytes] eq $bit)) {
	    $readbytes++;$count++;
	    if ($readbytes == $bufflen) {
		return END_OF_FILE if fillbuffer($readbytes) < 0;
		$start-=(1024*1024);
	    }
	}
	last if $count > 2478;		# if "marker" is too long for 400 baud
	$temptimer=$readbytes;
	while (($abuff[$temptimer] ne $bit)) {
	    $temptimer++;
	    if ($temptimer == $bufflen) {
		last if fillbuffer($temptimer) < 0;
		$start-=(1024*1024);
	    }
	}
	if (($temptimer-$readbytes) < MINBITLEN) { # !$bit was most probably a random spike, skip it
	    $count+=($temptimer-$readbytes);
	    $readbytes=$temptimer;
	    $i--;					#continue looking for the same bit
	    print $LOGFILE "SPIKE in marker\n";
	 }
    }
    if (($count > 2478) or ($count < 800)) {	# if "marker" is too long for 400 or too short for 1200 baud
	$readbytes-=$count;			# rewind
	$bitlen=74 if $bitlen==-1;		# (set standard $bitlen if there was no marker previously)
	$minbitlen=int($bitlen/2.5); 		# (bit should never be shorter than this)
	return -3;				# it can not be a marker
    }						# continue with old $bitlen value

    $savedbitlen=$bitlen;
    $bitlen=int($count/20+0.5);			# computed bit length
    $bitlen=int(($bitlen+$savedbitlen)/2+0.5)	# don't allow too big changes in baud rate, but only when:
	if (($savedbitlen!=-1) 			# bitlen was previously found
	and (abs($bitlen-$savedbitlen) > 2) 	# and difference is grater than 2
	and ($irgtimer<100000));		# and IRG was long enough (indicating next part of the tape)
    $minbitlen=int($bitlen/2.5); 		# bit should never be shorter than this

    # verify, whether there are really two 0x55 bytes at the beginning
#    $temptimer=$readbytes;
#    $readbytes=$start;
#    $result=getbyte(my $databyte1, my $casdatabyte1);
#    $result=getbyte(my $databyte2, my $casdatabyte2);
#    if ($databyte1 ne "85" or $databyte2 ne "85") { # if any of two first bytes is not equal to 0x55
#	print $LOGFILE "$databyte1,$databyte2\n";
#	$readbytes=$start;			# rewind
#	$bitlen=$savedbitlen;
#	$bitlen=74 if $bitlen==-1;		# (set standard $bitlen if there was no marker previously)
#	$minbitlen=int($bitlen/2.5); 		# (bit should never be shorter than this)
#	return -3;				# because it can not be a marker
#    }

#    $readbytes=$temptimer;			# Everythig OK, continue

    return 1;	#OK
}

sub skipirg {

    my $temptimer;

    $irgtimer=0;
    # skip until mark signal
    while ($abuff[$readbytes] ne MARK) {
	$readbytes++;
	if ($readbytes == $bufflen) {
	    return END_OF_FILE if fillbuffer($readbytes) < 0;
	}
    }
    # search for a start bit
    while(true) {
	while (($abuff[$readbytes] eq MARK)) {
	    $irgtimer++; $readbytes++;
	    if ($readbytes == $bufflen) {
		return END_OF_FILE if fillbuffer($readbytes) < 0;
	    }
	}
	$temptimer=$readbytes;
	while (($abuff[$temptimer] eq START)) {
	    $temptimer++;
	    if ($temptimer == $bufflen) {
		return END_OF_FILE if fillbuffer($temptimer) < 0;
	    }
	}
	if ($temptimer-$readbytes > $minbitlen) {	# found a start bit after IRG
	    last; }
	else {
	    $irgtimer+=$temptimer-$readbytes;
	    $readbytes=$temptimer;			# found a random spike, go on
	}
    }
    return $irgtimer;
}

sub getrecord {

# get one record of bytes

    $reclen=0;
    my $databyte;
    my $result;
    my $casdatabyte;

    @record=();
    @casrecord=();
    $result=skipirg();
    return $result if $result==END_OF_FILE;	# EOF
    print $LOGFILE "IRG: " . $irgtimer . "\n";
    $result=readmarker();
    return $result if (($result==END_OF_FILE) or ($result==NOT_GOOD));	# EOF or ERROR
    if ($result==-3) {
	print $LOGFILE "Non-standard record!\n" ;  # No marker, protected tape etc.
    } else {
	$record[$reclen]='85';
	$casrecord[$reclen++]=chr(85);
	$record[$reclen]='85';
	$casrecord[$reclen++]=chr(85); # 2 marker bytes
    }
    print $LOGFILE "Bit length: $bitlen samples\n";
    while (true) {
	$longzero=0;
	$result=getbyte($databyte,$casdatabyte);
	return ZERO_DETECTED if $result==ZERO_DETECTED;
	$record[$reclen]=$databyte;
	$casrecord[$reclen]=$casdatabyte;
	$reclen++;
	return $result if $result==END_OF_FILE;	# EOF
	last if $result==IRG_DETECTED;	# IRG
    }
    return $reclen;	#OK
}

sub controlsum {

# compute checksum of data block

    my $sum=0;

    foreach (@record[0..$#record-1]) {
#	$sum=($sum+$_) % 255;
	$sum+=$_;
	$sum-=255 if $sum>255;
    }
return $sum;
}

########################
### START PROCESSING ###
########################

#binmode(STDIN); # for windows only ?
#binmode(STDOUT); # for windows only ?

$totalbytes=0;			# total number of input bytes
$readbytes=44;			# number of bytes read so far (skip wav header)
$bufflen=0;			# buffer is initially empty
$bitlen=-1;			# length of bit (in samples)
$minbitlen=int($bitlen/2.5);	# bit should never be shorter than this

my $i;				# just loop index
my $records=0;			# blocks counter
my $cksum;			# checksum
my $status;
my $errors=0;			# checksum error counter
my $longzeros=0;		# number of "long zeros"
my $nonstd=0;			# number of non-standard data records

my $samplerate;			# samplarate XD
my $title="";			# tape name
my $help=0;
my $man=0;
my $skipshorter=0;

my $lastbitlen=-1;	# remember previous bit length

GetOptions('title|t=s' => \$title,
	    'skipshorter=i' => \$skipshorter,
	    'help|h' => \$help,
	    'man' => \$man) or pod2usage(2);

pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

pod2usage(1) if !@ARGV;


open($INFILE,"<" . @ARGV[0]) or die "Can't open input file $ARGV[0]";

$ARGV[0]=~s/(.*)(-squared\.wav$)/$1/;

open($CASFILE,">" . @ARGV[0] . ".cas") or die "Can't open output cas file.";
open($LOGFILE,">" . @ARGV[0] . ".log") or die "Can't open output log file.";

# uncomment for a bit of debugging
# $LOGFILE->autoflush(1);

@abuff=();
@record=();
@casrecord=();

fillbuffer() or die "Can't read data from input file" ;

# start writting to ,cas file
print $CASFILE "FUJI",pack("v",length($title)),pack("v",0),$title;

# get sample rate from wav file
$samplerate=unpack("V",join('',@abuff[24..27]));

# main loop
while (true) {
    print "Data blocks: $records, checksum errors: $errors, non-standard data blocks: $nonstd, long zeros: $longzeros.\r";
    STDOUT->flush;
    $status=getrecord();
    last if (($status==END_OF_FILE) and ($reclen==0));
    if ($reclen!=0) {
	$cksum=controlsum();
	$records++;
	print $LOGFILE "record: $records, bytes: $reclen\n";
	print $LOGFILE join(',',@record);
	print $LOGFILE "\ncksum: " . $cksum . ", from record: " . @record[$#record] . "\n";
	if ($cksum != @record[$#record]) {
	    print $LOGFILE "CHECKSUM ERROR in ",
		$reclen==132 ? "standard" : "non-standard",
		" record $records!\n" ;
	    $errors++;
	}
	$nonstd++ if $reclen!=132;
	if ($reclen>=$skipshorter) {
	    if ($bitlen != $lastbitlen) {		# store baud rate whenever it changes
		print $CASFILE "baud",pack("v",0),
		    pack("v",int((int(($samplerate/44100)*73.5+0.5)/$bitlen)*600+0.5));
		$lastbitlen=$bitlen;
	    }
		print $CASFILE "data",pack("v",$reclen),pack("v",int($irgtimer/$samplerate*1000));	# record length and irg
		print $CASFILE join('',@casrecord);							# data
	} else {
	    print $LOGFILE "skipped short ($reclen bytes) record\n" if $reclen!=0;
#	    print $CASFILE "data",pack("v",0),pack("v",int($irgtimer/$samplerate*1000));	# record length and irg
	}
    } 
    if ($status==ZERO_DETECTED) {
	print $LOGFILE "Long space signal ($longzero samples).\n";
	print $CASFILE "fsk ",pack("v",2),pack("v",int($irgtimer/$samplerate*1000)),pack("v",int($longzero/$samplerate*10000));
	$longzeros++;
    }
    last if (($status==END_OF_FILE) or ($status==NOT_GOOD));
}

print "SUMMARY: Data blocks: $records, checksum errors: $errors,",
	" non-standard data blocks: $nonstd, long zeros: $longzeros, read $totalbytes input bytes.\n";
STDOUT->flush;

print $LOGFILE "SUMMARY: Records: $records, checksum errors: $errors,",
	" non-standard data blocks: $nonstd, long zeros: $longzeros, read $totalbytes input bytes.\n";

close($CASFILE);
close($INFILE);
close $LOGFILE;

1;	#exit nicely

__END__

=head1 NAME

square2cas.pl - converting square wave to .cas file

=head1 SYNOPSIS

square2cas.pl [options] file.wav

  Options (default value):
    --title | -t <title>	tape title (none)
    --skipshorter <bytes>	skip records shorter than this

    --help | -h			this brief help
    --man			full documentation

=head1 OPTIONS

=over 8

=item B<--help | -h>

Print a brief help message and exits.

=item B<--man>

Prints the manual page and exits.

=item B<--title | -t <title>>

Specifies the title of the program, stored in the .cas file.

=item B<--skipshorter <bytes>>

Use it if You want to skip obviously wrong, short data records,
don't store them in .cas file.

=back

=head1 DESCRIPTION

B<square2cas.pl> will convert square wave in .wav file into tape image in .cas format.
Written by: B<FUJI> in 2009
Nick 'FUJI' was registered at atari.area (http://atariarea.krap.pl/) in 2009

=cut
