#!/bin/bash

# Square wave to Atari tape converter
# Written by FUJI in 2009
# Nick 'FUJI' was registered at atari.area (http://atariarea.krap.pl/) in 2009

# This script uses the following software:
#  ecasound (http://ecasound.seul.org/ecasound/index.php)
#  (LADSPA) CMT plugins (http://www.ladspa.org/cmt/)
#  (LADSPA) Omins plugins (http://www.nongnu.org/om-synth/omins.html#plugins)

[ "$#" -eq "0" ] && { echo "Parameters are missing..."; exit 1; }

while [ "$#" -ne "1" ]; do
    case $1 in
# 'mark' signal frequency, default shoud be OK
    "-mark") { shift; MARKFREQ=$1; shift; }
	;;
# 'space' signal frequency, default shoud be OK
    "-space") { shift; SPACEFREQ=$1; shift; }
	;;
# amplitude of output signal, ranging from 0 to 1
    "-gain") { shift; GAIN=$1; shift; }
	;;
# ignore the rest until one remains
    *) { shift; }
	;;
    esac
done

[ "$#" -eq "0" ] && { echo "Parameters are BAD..."; exit 1; }

INFILE=${1}

MARKFREQ=${MARKFREQ:-5327}
SPACEFREQ=${SPACEFREQ:-3995}
GAIN=${GAIN:-0.5}

OUTFILE="${INFILE%-rebuilt.*}-mastered.wav"

RATE=`ecalength -s -f -r ${1} 2>/dev/null`

# I don't want unneded error messages
exec 2>/dev/null
export MALLOC_CHECK_=0

echo -e "Processing...\c"

ecasound -x -q -i:${INFILE} \
    -f:s16_le,1,${RATE} \
    -el:range_trans_cr,-1,1,${SPACEFREQ},${MARKFREQ} -el:sine_faac,${GAIN} \
    -f:s16_le,2,${RATE} -chmove:1,2 -o ${OUTFILE}

if [ "$?" -eq "0" ]; then
    echo " finished."
else
    echo " error!"
fi
